@@ -1,12 +1,22 @@
 from __future__ import annotations
-from .exceptions import AttributeMissingError, ColumnNotFoundError
+
+from typing import Any
+from typing import Dict
+from typing import List
+from typing import Optional
+from typing import Tuple
+from typing import Union
+
+from .exceptions import AttributeMissingError
+from .exceptions import ColumnNotFoundError
+from .exceptions import DuplicateReferenceError
 
 
 class SQLOjbect:
     '''
     Base class for all SQL objects.
     '''
-    required_attributes = []
+    required_attributes: Tuple[str, ...] = ()
 
     def check_attributes_for_sql(self):
         '''
@@ -19,6 +29,22 @@ class SQLOjbect:
                     f'Cannot render SQL. Missing required attribute "{attr}".'
                 )
 
+    def __setattr__(self, name: str, value: Any):
+        """
+        Required for type testing with MyPy.
+        """
+        super().__setattr__(name, value)
+
+    def __eq__(self, other: object) -> bool:
+        """
+        Two instances of the same SQLObject subclass are equal if all their
+        attributes are equal.
+        """
+
+        if isinstance(other, self.__class__):
+            return self.__dict__ == other.__dict__
+        return False
+
 
 class ReferenceBlueprint:
     '''
@@ -32,14 +58,14 @@ class ReferenceBlueprint:
 
     def __init__(self,
                  type_: str,
-                 name: str = None,
-                 table1: str = None,
-                 col1: str = None,
-                 table2: str = None,
-                 col2: str = None,
-                 comment: str = None,
-                 on_update: str or None = None,
-                 on_delete: str or None = None):
+                 name: Optional[str] = None,
+                 table1: Optional[str] = None,
+                 col1: Optional[str] = None,
+                 table2: Optional[str] = None,
+                 col2: Optional[str] = None,
+                 comment: Optional[str] = None,
+                 on_update: Optional[str] = None,
+                 on_delete: Optional[str] = None):
         self.type = type_
         self.name = name if name else None
         self.table1 = table1 if table1 else None
@@ -51,21 +77,21 @@ class ReferenceBlueprint:
         self.on_delete = on_delete
 
     def __repr__(self):
-        components = [f"ReferenceBlueprint({repr(self.type)}"]
+        components = [f"ReferenceBlueprint({self.type!r}"]
         if self.name:
-            components.append(f'name={repr(self.name)}')
+            components.append(f'name={self.name!r}')
         if self.table1:
-            components.append(f'table1={repr(self.table1)}')
+            components.append(f'table1={self.table1!r}')
         if self.col1:
-            components.append(f'col1={repr(self.col1)}')
+            components.append(f'col1={self.col1!r}')
         if self.table2:
-            components.append(f'table2={repr(self.table2)}')
+            components.append(f'table2={self.table2!r}')
         if self.col2:
-            components.append(f'col2={repr(self.col2)}')
+            components.append(f'col2={self.col2!r}')
         if self.on_update:
-            components.append(f'on_update={repr(self.on_update)}')
+            components.append(f'on_update={self.on_update!r}')
         if self.on_delete:
-            components.append(f'on_delete={repr(self.on_delete)}')
+            components.append(f'on_delete={self.on_delete!r}')
         return ', '.join(components) + ')'
 
     def __str__(self):
@@ -88,7 +114,7 @@ class Reference(SQLOjbect):
     It is a separate object, which is not connected to Table or Column objects
     and its `sql` property contains the ALTER TABLE clause.
     '''
-    required_attributes = ['type', 'table1', 'col1', 'table2', 'col2']
+    required_attributes = ('type', 'table1', 'col1', 'table2', 'col2')
 
     ONE_TO_MANY = '<'
     MANY_TO_ONE = '>'
@@ -97,18 +123,18 @@ class Reference(SQLOjbect):
     def __init__(self,
                  type_: str,
                  table1: Table,
-                 col1: Column,
+                 col1: Union[Column, List[Column]],
                  table2: Table,
-                 col2: Column,
-                 name: str = None,
-                 comment: str = None,
-                 on_update: str or None = None,
-                 on_delete: str or None = None):
+                 col2: Union[Column, List[Column]],
+                 name: Optional[str] = None,
+                 comment: Optional[str] = None,
+                 on_update: Optional[str] = None,
+                 on_delete: Optional[str] = None):
         self.type = type_
         self.table1 = table1
-        self.col1 = col1
+        self.col1 = [col1] if isinstance(col1, Column) else list(col1)
         self.table2 = table2
-        self.col2 = col2
+        self.col2 = [col2] if isinstance(col2, Column) else list(col2)
         self.name = name if name else None
         self.comment = comment
         self.on_update = on_update
@@ -116,18 +142,18 @@ class Reference(SQLOjbect):
 
     def __repr__(self):
         components = [
-            f"Reference({repr(self.type)}",
-            f"{repr(self.table1)}",
-            f"{repr(self.col1)}",
-            f"{repr(self.table2)}",
-            f"{repr(self.col2)}",
+            f"Reference({self.type!r}",
+            f"{self.table1!r}",
+            f"{self.col1!r}",
+            f"{self.table2!r}",
+            f"{self.col2!r}",
         ]
         if self.name:
-            components.append(f'name={repr(self.name)}')
+            components.append(f'name={self.name!r}')
         if self.on_update:
-            components.append(f'on_update={repr(self.on_update)}')
+            components.append(f'on_update={self.on_update!r}')
         if self.on_delete:
-            components.append(f'on_delete={repr(self.on_delete)}')
+            components.append(f'on_delete={self.on_delete!r}')
         return ', '.join(components) + ')'
 
     def __str__(self):
@@ -135,12 +161,12 @@ class Reference(SQLOjbect):
         if self.table1:
             components.append(self.table1.name)
         if self.col1:
-            components.append(f'.{self.col1.name}')
+            components.append(f'[{", ".join(c.name for c in self.col1)}]')
         components.append(f' {self.type} ')
         if self.table2:
             components.append(self.table2.name)
         if self.col2:
-            components.append(f'.{self.col2.name}')
+            components.append(f'[{", ".join(c.name for c in self.col2)}]')
         return ''.join(components) + ')'
 
     @property
@@ -156,14 +182,14 @@ class Reference(SQLOjbect):
 
         if self.type in (self.MANY_TO_ONE, self.ONE_TO_ONE):
             t1 = self.table1
-            c1 = self.col1
+            c1 = ', '.join(self.col1)
             t2 = self.table2
-            c2 = self.col2
+            c2 = ', '.join(self.col2)
         else:
             t1 = self.table2
-            c1 = self.col2
+            c1 = ', '.join(self.col2)
             t2 = self.table1
-            c2 = self.col1
+            c2 = ', '.join(self.col1)
 
         result = (
             f'ALTER TABLE "{t1.name}" ADD {c}FOREIGN KEY ("{c1.name}") '
@@ -182,34 +208,36 @@ class TableReference(SQLOjbect):
     This object should be assigned to the `refs` attribute of a Table object.
     Its `sql` property contains the inline definition of the FOREIGN KEY clause.
     '''
-    required_attributes = ['col', 'ref_table', 'ref_col']
+    required_attributes = ('col', 'ref_table', 'ref_col')
 
     def __init__(self,
-                 col: Column,
+                 col: Union[Column, List[Column]],
                  ref_table: Table,
-                 ref_col: Column,
-                 name: str or None = None,
-                 on_delete: str or None = None,
-                 on_update: str or None = None):
-        self.col = col
+                 ref_col: Union[Column, List[Column]],
+                 name: Optional[str] = None,
+                 on_delete: Optional[str] = None,
+                 on_update: Optional[str] = None):
+        self.col = [col] if isinstance(col, Column) else list(col)
         self.ref_table = ref_table
-        self.ref_col = ref_col
+        self.ref_col = [ref_col] if isinstance(ref_col, Column) else list(ref_col)
         self.name = name
         self.on_update = on_update
         self.on_delete = on_delete
 
     def __repr__(self):
-        components = [f"TableReference({repr(self.col)}, {repr(self.ref_table)}, {repr(self.ref_col)}"]
+        components = [f"TableReference({self.col!r}, {self.ref_table!r}, {self.ref_col!r}"]
         if self.name:
-            components.append(f'name={repr(self.name)}')
+            components.append(f'name={self.name!r}')
         if self.on_update:
-            components.append(f'on_update={repr(self.on_update)}')
+            components.append(f'on_update={self.on_update!r}')
         if self.on_delete:
-            components.append(f'on_delete={repr(self.on_delete)}')
+            components.append(f'on_delete={self.on_delete!r}')
         return ', '.join(components) + ')'
 
     def __str__(self):
-        return f"TableReference({self.col.name} -> {self.ref_table.name}.{self.ref_col.name})"
+        cols = '", "'.join(c.name for c in self.col)
+        ref_cols = '", "'.join(c.name for c in self.ref_col)
+        return f"TableReference({cols} -> {self.ref_table.name}[{ref_cols}])"
 
     @property
     def sql(self):
@@ -221,9 +249,11 @@ class TableReference(SQLOjbect):
         '''
         self.check_attributes_for_sql()
         c = f'CONSTRAINT "{self.name}" ' if self.name else ''
+        cols = '", "'.join(c.name for c in self.col)
+        ref_cols = '", "'.join(c.name for c in self.ref_col)
         result = (
-            f'{c}FOREIGN KEY ("{self.col.name}") '
-            f'REFERENCES "{self.ref_table.name} ("{self.ref_col.name}")'
+            f'{c}FOREIGN KEY ("{cols}") '
+            f'REFERENCES "{self.ref_table.name} ("{ref_cols}")'
         )
         if self.on_update:
             result += f' ON UPDATE {self.on_update.upper()}'
@@ -245,11 +275,18 @@ class Note:
     def __repr__(self):
         return f'Note({repr(self.text)})'
 
+    @property
+    def sql(self):
+        if self.text:
+            return '\n'.join(f'-- {line}' for line in self.text.split('\n'))
+        else:
+            return ''
+
 
 class Column(SQLOjbect):
     '''Class representing table column.'''
 
-    required_attributes = ['name', 'type']
+    required_attributes = ('name', 'type')
 
     def __init__(self,
                  name: str,
@@ -258,10 +295,10 @@ class Column(SQLOjbect):
                  not_null: bool = False,
                  pk: bool = False,
                  autoinc: bool = False,
-                 default=None,
-                 note: str or None = None,
-                 ref_blueprints: list = None,
-                 comment: str or None = None):
+                 default: Optional[Union[str, int, bool, float]] = None,
+                 note: Optional[Note] = None,
+                 ref_blueprints: Optional[List[ReferenceBlueprint]] = None,
+                 comment: Optional[str] = None):
         self.name = name
         self.type = type_
         self.unique = unique
@@ -277,10 +314,10 @@ class Column(SQLOjbect):
         for ref in self.ref_blueprints:
             ref.col1 = self.name
 
-        self._table = None
+        self._table: Optional[Table] = None
 
     @property
-    def table(self):
+    def table(self) -> Optional[Table]:
         return self._table
 
     @table.setter
@@ -309,10 +346,12 @@ class Column(SQLOjbect):
             components.append('NOT NULL')
         if self.default is not None:
             components.append('DEFAULT ' + str(self.default))
+        if self.note:
+            components.append(self.note.sql)
         return ' '.join(components)
 
     def __repr__(self):
-        components = [f"Column({repr(self.name)}, {repr(self.type)}"]
+        components = [f"Column({self.name!r}, {self.type!r}"]
         if self.unique:
             components.append(f'unique=True')
         if self.not_null:
@@ -322,9 +361,9 @@ class Column(SQLOjbect):
         if self.autoinc:
             components.append(f'autoinc=True')
         if self.default:
-            components.append(f'default={repr(self.default)}')
+            components.append(f'default={self.default!r}')
         if self.note:
-            components.append(f'note={repr(self.note)}')
+            components.append(f'note={self.note!r}')
         return ', '.join(components) + ')'
 
     def __str__(self):
@@ -338,25 +377,25 @@ class Column(SQLOjbect):
         if self.autoinc:
             components.append(f'autoincrement')
         if self.default:
-            components.append(f'default={repr(self.default)}')
+            components.append(f'default={self.default!r}')
         return ' '.join(components) + ')'
 
 
 class Index(SQLOjbect):
     '''Class representing index.'''
-    required_attributes = ['subjects']
+    required_attributes = ('subjects', 'table')
 
     def __init__(self,
-                 subject_names: list,
-                 name: str or None = None,
-                 table: Table or None = None,
+                 subject_names: List[str],
+                 name: Optional[str] = None,
+                 table: Optional[Table] = None,
                  unique: bool = False,
-                 type_: str or None = None,
+                 type_: Optional[str] = None,
                  pk: bool = False,
-                 note: Note or None = None,
-                 comment: str or None = None):
+                 note: Optional[Note] = None,
+                 comment: Optional[str] = None):
         self.subject_names = subject_names
-        self.subjects = []
+        self.subjects: List[Union[Column, str]] = []
 
         self.name = name if name else None
         self.table = table
@@ -367,17 +406,17 @@ class Index(SQLOjbect):
         self.comment = comment
 
     def __repr__(self):
-        components = [f"Index({repr(self.subject_names)}"]
+        components = [f"Index({self.subject_names!r}"]
         if self.name:
-            components.append(f'name={repr(self.name)}')
+            components.append(f'name={self.name!r}')
         if self.unique:
             components.append(f'unique=True')
         if self.type:
-            components.append(f'type_={repr(self.type)}')
+            components.append(f'type_={self.type!r}')
         if self.pk:
             components.append(f'pk=True')
         if self.note:
-            components.append(f'note_={repr(self.note)}')
+            components.append(f'note_={self.note!r}')
         return ', '.join(components) + ')'
 
     def __str__(self):
@@ -387,7 +426,7 @@ class Index(SQLOjbect):
         else:
             components[0] += '(' + ', '.join(self.subjects) + ')'
         if self.name:
-            components.append(f'{repr(self.name)}')
+            components.append(f'{self.name!r}')
         if self.unique:
             components.append(f'unique')
         if self.type:
@@ -411,7 +450,7 @@ class Index(SQLOjbect):
 
         '''
         self.check_attributes_for_sql()
-        keys = ', '.join(f'"{key.name}"' for key in self.subjects)
+        keys = ', '.join(f'"{key.name}"' if isinstance(key, Column) else key for key in self.subjects)
         if self.pk:
             return f'PRIMARY KEY ({keys})'
 
@@ -425,32 +464,35 @@ class Index(SQLOjbect):
         if self.type:
             components.append(f'USING {self.type.upper()}')
         components.append(f'({keys})')
-        return ' '.join(components) + ';'
+        result = ' '.join(components) + ';'
+        if self.note:
+            result += f' {self.note.sql}'
+        return result
 
 
 class Table(SQLOjbect):
     '''Class representing table.'''
 
-    required_attributes = ['name']
+    required_attributes = ('name',)
 
     def __init__(self,
                  name: str,
-                 alias: str or None = None,
-                 note: Note or None = None,
-                 header_color: str or None = None,
-                 refs: list or None = None,
-                 comment: str or None = None):
+                 alias: Optional[str] = None,
+                 note: Optional[Note] = None,
+                 header_color: Optional[str] = None,
+                 refs: Optional[List[TableReference]] = None,
+                 comment: Optional[str] = None):
         self.name = name
-        self.columns = []
-        self.indexes = []
-        self.column_dict = {}
+        self.columns: List[Column] = []
+        self.indexes: List[Index] = []
+        self.column_dict: Dict[str, Column] = {}
         self.alias = alias if alias else None
         self.note = note or Note('')
         self.header_color = header_color
         self.refs = refs or []
         self.comment = comment
 
-    def add_column(self, c: Column):
+    def add_column(self, c: Column) -> None:
         '''
         Adds column to self.columns attribute and sets in this column the
         `table` attribute.
@@ -459,21 +501,35 @@ class Table(SQLOjbect):
         self.columns.append(c)
         self.column_dict[c.name] = c
 
-    def add_index(self, i: Index):
+    def add_index(self, i: Index) -> None:
         '''
         Adds index to self.indexes attribute and sets in this index the
         `table` attribute.
         '''
         for subj in i.subject_names:
-            col = self.get(subj)
-            if not col:
-                raise ColumnNotFoundError(f'Cannot add index, column "{subj}" not defined in table "{self.name}".')
-            i.subjects.append(col)
+            if subj.startswith('(') and subj.endswith(')'):
+                # subject is an expression, add it as string
+                i.subjects.append(subj)
+            else:
+                try:
+                    col = self[subj]
+                    i.subjects.append(col)
+                except KeyError:
+                    raise ColumnNotFoundError(f'Cannot add index, column "{subj}" not defined in table "{self.name}".')
 
         i.table = self
         self.indexes.append(i)
 
-    def __getitem__(self, k: int or str) -> Table:
+    def add_ref(self, r: TableReference) -> None:
+        '''
+        Adds a reference to the table. If reference already present in the table,
+        raises DuplicateReferenceError.
+        '''
+        if r in self.refs:
+            raise DuplicateReferenceError(f'Reference with same endpoints {r} already present in the table.')
+        self.refs.append(r)
+
+    def __getitem__(self, k: Union[int, str]) -> Column:
         if isinstance(k, int):
             return self.columns[k]
         else:
@@ -486,15 +542,15 @@ class Table(SQLOjbect):
         return iter(self.columns)
 
     def __repr__(self):
-        components = [f"Table({repr(self.name)}, {repr(self.columns)}"]
+        components = [f"Table({self.name!r}, {self.columns!r}"]
         if self.alias:
-            components.append(f'alias={repr(self.alias)}')
+            components.append(f'alias={self.alias!r}')
         if self.indexes:
-            components.append(f'indexes={repr(self.indexes)}')
+            components.append(f'indexes={self.indexes!r}')
         if self.note:
-            components.append(f'note={repr(self.note)}')
+            components.append(f'note={self.note!r}')
         if self.header_color:
-            components.append(f'header_color={repr(self.header_color)}')
+            components.append(f'header_color={self.header_color!r}')
         return ', '.join(components) + ')'
 
     def __str__(self):
@@ -521,6 +577,8 @@ class Table(SQLOjbect):
         '''
         self.check_attributes_for_sql()
         components = [f'CREATE TABLE "{self.name}" (']
+        if self.note:
+            components.append(f'  {self.note.sql}')
         body = []
         body.extend('  ' + c.sql for c in self.columns)
         body.extend('  ' + i.sql for i in self.indexes if i.pk)
@@ -536,29 +594,36 @@ class EnumItem:
 
     def __init__(self,
                  name: str,
-                 note: Note or None = None,
-                 comment: str or None = None):
+                 note: Optional[Note] = None,
+                 comment: Optional[str] = None):
         self.name = name
         self.note = note or Note('')
         self.comment = comment
 
     def __repr__(self):
-        components = [f'EnumItem({repr(self.name)}']
+        components = [f'EnumItem({self.name!r}']
         if self.note:
-            components.append(f'note={repr(self.note)}')
+            components.append(f'note={self.note!r}')
         return ', '.join(components) + ')'
 
     def __str__(self):
         return self.name
 
+    @property
+    def sql(self):
+        components = [f"'{self.name}',"]
+        if self.note:
+            components.append(self.note.sql)
+        return ' '.join(components)
+
 
 class Enum(SQLOjbect):
-    required_attributes = ['name', 'items']
+    required_attributes = ('name', 'items')
 
     def __init__(self,
                  name: str,
-                 items: list,
-                 comment: str or None = None):
+                 items: List[EnumItem],
+                 comment: Optional[str] = None):
         self.name = name
         self.items = items
         self.comment = comment
@@ -566,20 +631,18 @@ class Enum(SQLOjbect):
     def get_type(self):
         return EnumType(self.name, self.items)
 
-    def __getitem__(self, key):
+    def __getitem__(self, key) -> EnumItem:
         return self.items[key]
 
     def __iter__(self):
         return iter(self.items)
 
     def __repr__(self):
-        return f'Enum({repr(self.name)}, {repr(self.items)})'
+        return f'Enum({self.name!r}, {self.items!r})'
 
     def __str__(self):
         return (
-            f'Enum {self.name} (' +
-            ', '.join(str(i) for i in self.items) +
-            ')'
+            f'Enum {self.name} (' + ', '.join(str(i) for i in self.items) + ')'
         )
 
     @property
@@ -596,11 +659,8 @@ class Enum(SQLOjbect):
 
         '''
         self.check_attributes_for_sql()
-        items = []
-        for item in self.items:
-            items.append(f"  '{item}',")
         return f'CREATE TYPE "{self.name}" AS ENUM (\n' +\
-               '\n'.join(items) +\
+               '\n'.join(f'  {i.sql}' for i in self.items) +\
                '\n);'
 
 
@@ -610,25 +670,31 @@ class EnumType(Enum):
     '''
 
     def __repr__(self):
-        return f'EnumType({repr(self.name)}, {repr(self.items)})'
+        return f'EnumType({self.name!r}, {self.items!r})'
 
     def __str__(self):
         return self.name
 
 
 class TableGroup:
+    '''
+    TableGroup `items` parameter initially holds just the names of the tables,
+    but after parsing the whole document, PyDBMLParseResults class replaces
+    them with references to actual tables.
+    '''
+
     def __init__(self,
                  name: str,
-                 items: list,
-                 comment: str or None = None):
+                 items: Union[List[str], List[Table]],
+                 comment: Optional[str] = None):
         self.name = name
         self.items = items
         self.comment = comment
 
     def __repr__(self):
-        return f'TableGroup({repr(self.name)}, {repr(self.items)})'
+        return f'TableGroup({self.name!r}, {self.items!r})'
 
-    def __getitem__(self, key):
+    def __getitem__(self, key) -> str:
         return self.items[key]
 
     def __iter__(self):
@@ -638,18 +704,18 @@ class TableGroup:
 class Project:
     def __init__(self,
                  name: str,
-                 items: dict or None = None,
-                 note: Note or None = None,
-                 comment: str or None = None):
+                 items: Optional[Dict[str, str]] = None,
+                 note: Optional[Note] = None,
+                 comment: Optional[str] = None):
         self.name = name
         self.items = items
         self.note = note or Note('')
         self.comment = comment
 
     def __repr__(self):
-        components = [f'Project({repr(self.name)}']
+        components = [f'Project({self.name!r}']
         if self.items:
-            components.append(f'items={repr(self.items)}')
+            components.append(f'items={self.items!r}')
         if self.note:
-            components.append(f'note={repr(self.note)}')
+            components.append(f'note={self.note!r}')
         return ', '.join(components) + ')'
