@@ -1,15 +1,31 @@
 from __future__ import annotations
+
 import pyparsing as pp
-from pathlib import PosixPath
+
 from io import TextIOWrapper
-from pydbml.definitions.table import table
-from pydbml.definitions.common import comment, _
-from pydbml.definitions.reference import ref
-from pydbml.definitions.enum import enum
-from pydbml.definitions.table_group import table_group
-from pydbml.definitions.project import project
-from pydbml.classes import Table, TableReference, Reference
-from pydbml.exceptions import TableNotFoundError, ColumnNotFoundError
+from pathlib import Path
+
+from typing import Dict
+from typing import List
+from typing import Optional
+from typing import Union
+
+from .classes import Enum
+from .classes import Project
+from .classes import Reference
+from .classes import ReferenceBlueprint
+from .classes import Table
+from .classes import TableGroup
+from .classes import TableReference
+from .definitions.common import _
+from .definitions.common import comment
+from .definitions.enum import enum
+from .definitions.project import project
+from .definitions.reference import ref
+from .definitions.table import table
+from .definitions.table_group import table_group
+from .exceptions import ColumnNotFoundError
+from .exceptions import TableNotFoundError
 
 pp.ParserElement.setDefaultWhitespaceChars(' \t\r')
 
@@ -32,12 +48,13 @@ class PyDBML:
     >>> from pathlib import Path
     >>> p = PyDBML(Path('schema.dbml'))
     '''
+
     def __new__(cls,
-                source_: str or PosixPath or TextIOWrapper or None = None):
+                source_: Optional[Union[str, Path, TextIOWrapper]] = None):
         if source_ is not None:
             if isinstance(source_, str):
                 source = source_
-            elif isinstance(source_, PosixPath):
+            elif isinstance(source_, Path):
                 with open(source_, encoding='utf8') as f:
                     source = f.read()
             else:  # TextIOWrapper
@@ -58,7 +75,7 @@ class PyDBML:
         return PyDBMLParseResults(text)
 
     @staticmethod
-    def parse_file(file: str or PosixPath or TextIOWrapper):
+    def parse_file(file: Union[str, Path, TextIOWrapper]):
         if isinstance(file, TextIOWrapper):
             source = file.read()
         else:
@@ -71,19 +88,20 @@ class PyDBML:
 
 class PyDBMLParseResults:
     def __init__(self, source: str):
-        self.tables = []
-        self.table_dict = {}
-        self.refs = []
-        self.ref_blueprints = []
-        self.enums = []
-        self.table_groups = []
-        self.project = None
+        self.tables: List[Table] = []
+        self.table_dict: Dict[str, Table] = {}
+        self.refs: List[Reference] = []
+        self.ref_blueprints: List[ReferenceBlueprint] = []
+        self.enums: List[Enum] = []
+        self.table_groups: List[TableGroup] = []
+        self.project: Optional[Project] = None
         self.source = source
 
         self._set_syntax()
         self._syntax.parseString(self.source, parseAll=True)
         self._validate()
         self._process_refs()
+        self._process_table_groups()
         self._set_enum_types()
 
     def __repr__(self):
@@ -111,7 +129,7 @@ class PyDBMLParseResults:
         )
         self._syntax = expr[...] + ('\n' | comment)[...] + pp.StringEnd()
 
-    def __getitem__(self, k: int or str) -> Table:
+    def __getitem__(self, k: Union[int, str]) -> Table:
         if isinstance(k, int):
             return self.tables[k]
         else:
@@ -150,22 +168,36 @@ class PyDBMLParseResults:
         Validate refs at the same time.
         '''
         for ref_ in self.ref_blueprints:
-            table1 = self.table_dict.get(ref_.table1)
-            if not table1:
+            for tb in self.tables:
+                if tb.name == ref_.table1 or tb.alias == ref_.table1:
+                    table1 = tb
+                    break
+            else:
                 raise TableNotFoundError('Error while parsing reference:'
                                          f'table "{ref_.table1}"" is not defined.')
-            table2 = self.table_dict.get(ref_.table2)
-            if not table2:
+            for tb in self.tables:
+                if tb.name == ref_.table2 or tb.alias == ref_.table2:
+                    table2 = tb
+                    break
+            else:
                 raise TableNotFoundError('Error while parsing reference:'
                                          f'table "{ref_.table2}"" is not defined.')
-            col1 = table1.get(ref_.col1)
-            if not col1:
-                raise ColumnNotFoundError('Error while parsing reference:'
-                                          f'column "{ref_.col1} not defined in table "{table1.name}".')
-            col2 = table2.get(ref_.col2)
-            if not col2:
-                raise ColumnNotFoundError('Error while parsing reference:'
-                                          f'column "{ref_.col2} not defined in table "{table2.name}".')
+            col1_names = [c.strip('() ') for c in ref_.col1.split(',')]
+            col1 = []
+            for col_name in col1_names:
+                try:
+                    col1.append(table1[col_name])
+                except KeyError:
+                    raise ColumnNotFoundError('Error while parsing reference:'
+                                              f'column "{col_name} not defined in table "{table1.name}".')
+            col2_names = [c.strip('() ') for c in ref_.col2.split(',')]
+            col2 = []
+            for col_name in col2_names:
+                try:
+                    col2.append(table2[col_name])
+                except KeyError:
+                    raise ColumnNotFoundError('Error while parsing reference:'
+                                              f'column "{col_name} not defined in table "{table2.name}".')
             self.refs.append(
                 Reference(
                     ref_.type,
@@ -175,8 +207,8 @@ class PyDBMLParseResults:
                     col2,
                     name=ref_.name,
                     comment=ref_.comment,
-                    on_update=ref_.on_delete,
-                    on_delete=ref_.on_update
+                    on_update=ref_.on_update,
+                    on_delete=ref_.on_delete
                 )
             )
 
@@ -200,7 +232,7 @@ class PyDBMLParseResults:
                     'on_update': ref_.on_update,
                     'on_delete': ref_.on_delete
                 }
-            table.refs.append(
+            table.add_ref(
                 TableReference(**init_dict)
             )
 
@@ -223,6 +255,13 @@ class PyDBMLParseResults:
                 if table_name not in self.table_dict:
                     raise TableNotFoundError(f'Cannot add Table Group "{tg.name}": table "{table_name}" not found.')
 
+    def _process_table_groups(self):
+        '''
+        Fill up each TableGroup's `item` attribute with references to actual tables.
+        '''
+        for tg in self.table_groups:
+            tg.items = [self[i] for i in tg.items]
+
     @property
     def sql(self):
         '''Returs SQL of the parsed results'''
