@@ -1,27 +1,50 @@
 from unittest import TestCase
-from pydbml.exceptions import AttributeMissingError, ColumnNotFoundError
-from pydbml.classes import (ReferenceBlueprint, SQLOjbect,
-                            TableReference, Column, Table, Index, Enum, EnumItem)
+
+from pydbml.classes import Column
+from pydbml.classes import Enum
+from pydbml.classes import EnumItem
+from pydbml.classes import Index
+from pydbml.classes import Note
+from pydbml.classes import ReferenceBlueprint
+from pydbml.classes import SQLOjbect
+from pydbml.classes import Table
+from pydbml.classes import TableReference
+from pydbml.exceptions import AttributeMissingError
+from pydbml.exceptions import ColumnNotFoundError
+from pydbml.exceptions import DuplicateReferenceError
 
 
-class TestDBMLObject:
-    def test_check_attributes_for_sql(self):
+class TestDBMLObject(TestCase):
+    def test_check_attributes_for_sql(self) -> None:
         o = SQLOjbect()
         o.a1 = None
         o.b1 = None
         o.c1 = None
-        o.required_attributes = ['a1', 'b1']
+        o.required_attributes = ('a1', 'b1')
         with self.assertRaises(AttributeMissingError):
             o.check_attributes_for_sql()
         o.a1 = 1
         with self.assertRaises(AttributeMissingError):
             o.check_attributes_for_sql()
-        o.a2 = 'a2'
+        o.b1 = 'a2'
         o.check_attributes_for_sql()
 
+    def test_comparison(self) -> None:
+        o1 = SQLOjbect()
+        o1.a1 = None
+        o1.b1 = 'c'
+        o1.c1 = 123
+        o2 = SQLOjbect()
+        o2.a1 = None
+        o2.b1 = 'c'
+        o2.c1 = 123
+        self.assertTrue(o1 == o2)
+        o1.a2 = True
+        self.assertFalse(o1 == o2)
+
 
 # class TestReferenceBlueprint(TestCase):
-#     def test_basic_sql(self):
+#     def test_basic_sql(self) -> None:
 #         r = ReferenceBlueprint(
 #             ReferenceBlueprint.MANY_TO_ONE,
 #             table1='bookings',
@@ -37,7 +60,7 @@ class TestDBMLObject:
 #         expected2 = 'ALTER TABLE "ids" ADD FOREIGN KEY ("id") REFERENCES "bookings ("country");'
 #         self.assertEqual(r.sql, expected2)
 
-#     def test_full(self):
+#     def test_full(self) -> None:
 #         r = ReferenceBlueprint(
 #             ReferenceBlueprint.MANY_TO_ONE,
 #             name='refname',
@@ -58,14 +81,14 @@ class TestDBMLObject:
 
 
 # class TestTableReference(TestCase):
-#     def test_basic_sql(self):
+#     def test_basic_sql(self) -> None:
 #         r = TableReference(col='order_id',
 #                            ref_table='orders',
 #                            ref_col='id')
 #         expected = 'FOREIGN KEY ("order_id") REFERENCES "orders ("id")'
 #         self.assertEqual(r.sql, expected)
 
-#     def test_full(self):
+#     def test_full(self) -> None:
 #         r = TableReference(col='order_id',
 #                            ref_table='orders',
 #                            ref_col='id',
@@ -77,13 +100,21 @@ class TestDBMLObject:
 
 
 class TestColumn(TestCase):
-    def test_basic_sql(self):
+    def test_basic_sql(self) -> None:
         r = Column(name='id',
                    type_='integer')
         expected = '"id" integer'
         self.assertEqual(r.sql, expected)
 
-    def test_pk_autoinc(self):
+    def test_note(self) -> None:
+        n = Note('Column note')
+        r = Column(name='id',
+                   type_='integer',
+                   note=n)
+        expected = '"id" integer -- Column note'
+        self.assertEqual(r.sql, expected)
+
+    def test_pk_autoinc(self) -> None:
         r = Column(name='id',
                    type_='integer',
                    pk=True,
@@ -91,7 +122,7 @@ class TestColumn(TestCase):
         expected = '"id" integer PRIMARY KEY AUTOINCREMENT'
         self.assertEqual(r.sql, expected)
 
-    def test_unique_not_null(self):
+    def test_unique_not_null(self) -> None:
         r = Column(name='id',
                    type_='integer',
                    unique=True,
@@ -99,14 +130,14 @@ class TestColumn(TestCase):
         expected = '"id" integer UNIQUE NOT NULL'
         self.assertEqual(r.sql, expected)
 
-    def test_default(self):
+    def test_default(self) -> None:
         r = Column(name='order',
                    type_='integer',
                    default=0)
         expected = '"order" integer DEFAULT 0'
         self.assertEqual(r.sql, expected)
 
-    def test_table_setter(self):
+    def test_table_setter(self) -> None:
         r1 = ReferenceBlueprint(
             ReferenceBlueprint.MANY_TO_ONE,
             name='refname',
@@ -137,7 +168,7 @@ class TestColumn(TestCase):
 
 
 class TestIndex(TestCase):
-    def test_basic_sql(self):
+    def test_basic_sql(self) -> None:
         t = Table('products')
         t.add_column(Column('id', 'integer'))
         r = Index(subject_names=['id'],
@@ -146,7 +177,18 @@ class TestIndex(TestCase):
         expected = 'CREATE INDEX ON "products" ("id");'
         self.assertEqual(r.sql, expected)
 
-    def test_unique_type_composite(self):
+    def test_note(self) -> None:
+        t = Table('products')
+        t.add_column(Column('id', 'integer'))
+        n = Note('Index note')
+        r = Index(subject_names=['id'],
+                  table=t,
+                  note=n)
+        t.add_index(r)
+        expected = 'CREATE INDEX ON "products" ("id"); -- Index note'
+        self.assertEqual(r.sql, expected)
+
+    def test_unique_type_composite(self) -> None:
         t = Table('products')
         t.add_column(Column('id', 'integer'))
         t.add_column(Column('name', 'varchar'))
@@ -158,7 +200,7 @@ class TestIndex(TestCase):
         expected = 'CREATE UNIQUE INDEX ON "products" USING HASH ("id", "name");'
         self.assertEqual(r.sql, expected)
 
-    def test_pk(self):
+    def test_pk(self) -> None:
         t = Table('products')
         t.add_column(Column('id', 'integer'))
         t.add_column(Column('name', 'varchar'))
@@ -169,16 +211,26 @@ class TestIndex(TestCase):
         expected = 'PRIMARY KEY ("id", "name")'
         self.assertEqual(r.sql, expected)
 
+    def test_composite_with_expression(self) -> None:
+        t = Table('products')
+        t.add_column(Column('id', 'integer'))
+        r = Index(subject_names=['id', '(id*3)'],
+                  table=t)
+        t.add_index(r)
+        self.assertEqual(r.subjects, [t['id'], '(id*3)'])
+        expected = 'CREATE INDEX ON "products" ("id", (id*3));'
+        self.assertEqual(r.sql, expected)
+
 
 class TestTable(TestCase):
-    def test_one_column(self):
+    def test_one_column(self) -> None:
         t = Table('products')
         c = Column('id', 'integer')
         t.add_column(c)
         expected = 'CREATE TABLE "products" (\n  "id" integer\n);\n'
         self.assertEqual(t.sql, expected)
 
-    def test_ref(self):
+    def test_ref(self) -> None:
         t = Table('products')
         c1 = Column('id', 'integer')
         c2 = Column('name', 'varchar2')
@@ -188,7 +240,7 @@ class TestTable(TestCase):
         c21 = Column('name_val', 'varchar2')
         t2.add_column(c21)
         r = TableReference(c2, t2, c21)
-        t.refs.append(r)
+        t.add_ref(r)
         expected = \
 '''CREATE TABLE "products" (
   "id" integer,
@@ -198,7 +250,31 @@ class TestTable(TestCase):
 '''
         self.assertEqual(t.sql, expected)
 
-    def test_ref_index(self):
+    def test_duplicate_ref(self) -> None:
+        t = Table('products')
+        c1 = Column('id', 'integer')
+        c2 = Column('name', 'varchar2')
+        t.add_column(c1)
+        t.add_column(c2)
+        t2 = Table('names')
+        c21 = Column('name_val', 'varchar2')
+        t2.add_column(c21)
+        r1 = TableReference(c2, t2, c21)
+        t.add_ref(r1)
+        r2 = TableReference(c2, t2, c21)
+        self.assertEqual(r1, r2)
+        with self.assertRaises(DuplicateReferenceError):
+            t.add_ref(r2)
+
+    def test_note(self) -> None:
+        n = Note('Table note')
+        t = Table('products', note=n)
+        c = Column('id', 'integer')
+        t.add_column(c)
+        expected = 'CREATE TABLE "products" (\n  -- Table note\n  "id" integer\n);\n'
+        self.assertEqual(t.sql, expected)
+
+    def test_ref_index(self) -> None:
         t = Table('products')
         c1 = Column('id', 'integer')
         c2 = Column('name', 'varchar2')
@@ -208,7 +284,7 @@ class TestTable(TestCase):
         c21 = Column('name_val', 'varchar2')
         t2.add_column(c21)
         r = TableReference(c2, t2, c21)
-        t.refs.append(r)
+        t.add_ref(r)
         i = Index(['id', 'name'])
         t.add_index(i)
         expected = \
@@ -222,7 +298,7 @@ CREATE INDEX ON "products" ("id", "name");
 '''
         self.assertEqual(t.sql, expected)
 
-    def test_index_inline(self):
+    def test_index_inline(self) -> None:
         t = Table('products')
         c1 = Column('id', 'integer')
         c2 = Column('name', 'varchar2')
@@ -239,7 +315,7 @@ CREATE INDEX ON "products" ("id", "name");
 '''
         self.assertEqual(t.sql, expected)
 
-    def test_add_column(self):
+    def test_add_column(self) -> None:
         t = Table('products')
         c1 = Column('id', 'integer')
         c2 = Column('name', 'varchar2')
@@ -249,7 +325,7 @@ CREATE INDEX ON "products" ("id", "name");
         self.assertEqual(c2.table, t)
         self.assertEqual(t.columns, [c1, c2])
 
-    def test_add_index(self):
+    def test_add_index(self) -> None:
         t = Table('products')
         c1 = Column('id', 'integer')
         c2 = Column('name', 'varchar2')
@@ -263,7 +339,7 @@ CREATE INDEX ON "products" ("id", "name");
         self.assertEqual(i2.table, t)
         self.assertEqual(t.indexes, [i1, i2])
 
-    def test_add_bad_index(self):
+    def test_add_bad_index(self) -> None:
         t = Table('products')
         c = Column('id', 'integer')
         i = Index(['id', 'name'])
@@ -273,7 +349,7 @@ CREATE INDEX ON "products" ("id", "name");
 
 
 class TestEnum(TestCase):
-    def test_simple_enum(self):
+    def test_simple_enum(self) -> None:
         items = [
             EnumItem('created'),
             EnumItem('running'),
@@ -287,5 +363,23 @@ class TestEnum(TestCase):
   'running',
   'donef',
   'failure',
+);'''
+        self.assertEqual(e.sql, expected)
+
+    def test_notes(self) -> None:
+        n = Note('EnumItem note')
+        items = [
+            EnumItem('created', note=n),
+            EnumItem('running'),
+            EnumItem('donef', note=n),
+            EnumItem('failure'),
+        ]
+        e = Enum('job_status', items)
+        expected = \
+'''CREATE TYPE "job_status" AS ENUM (
+  'created', -- EnumItem note
+  'running',
+  'donef', -- EnumItem note
+  'failure',
 );'''
         self.assertEqual(e.sql, expected)
